# Warren's answer to SysArch 121 Assignment 2.
#
# This file contains the library functions
# specified in the assignment.

	.text
###################################################################
#			     malloc()				  #
###################################################################

# This is a naive implementation of malloc()
# which simply wraps around the sbrk syscall.
# However, we do watch out for zero or negative
# arguments and return 0 (NULL) in this case.
#
# Arguments: $a0 is the amount to allocate in bytes
# Results: $v0 is the pointer to the allocation, or NULL
# if the allocation could not be made.
#
malloc:	subu $sp, $sp, 24	# Build the stack frame
	sw $ra, 20($sp)

	la $v0, 0		# Assume we need to return NULL
	blez $a0, mallend	# Return this if our argument is 0 or -ve

	li $v0, 9		# Otherwise, call sbrk
	syscall			# and return its result instead

mallend:
	lw $ra, 20($sp)		# Reload our return address
	addu $sp, $sp, 24	# Destroy our stack frame, and
	jr $ra			# return back to our caller


###################################################################
#			     strlen()				  #
###################################################################
#
# Determine and return the length of a string, not including the NUL.
# Arguments: $a0 is the pointer to the beginning of the string
# Results: $v0 is the length of the string

# Iterative version of strlen. Note the error checking
strlen:	subu $sp, $sp, 24       # Build the stack frame
        sw $ra, 20($sp)
        sw $t0, 0($sp)		# Save $t0 as we use it

	li $v0, 0		# Set count to zero
	beqz $a0, strlend	# Return 0 if we receive a NULL pointer

strloop:
	lb $t0, ($a0)		# Get char from string
	beqz $t0, strlend	# NUL char, so exit the loop
	addi $v0, $v0, 1	# count++
	addi $a0, $a0, 1	# Move pointer up to next char
	b strloop

strlend:
        lw $t0, 0($sp)		# Reload $t0
	lw $ra, 20($sp)		# Reload our return address
	addu $sp, $sp, 24	# Destroy our stack frame, and
	jr $ra			# return back to our caller


# Recursive version of strlen. Note the error checking
rstrlen: subu $sp, $sp, 24       # Build the stack frame
        sw $ra, 20($sp)
        sw $t0, 0($sp)		# Save $t0 as we use it

	beqz $a0, rstrlzero	# Return 0 if we receive a NULL pointer

	lb $t0, ($a0)		# Get char from string
	beqz $t0, rstrlzero	# NUL char, so return zero

	addi $a0, $a0, 1	# Move pointer up to next char
	jal rstrlen		# Recursively call ourselves
	addi $v0, $v0, 1	# and add 1 for our character
	b rstrlend		# then return the result
	
rstrlzero:
	li $v0, 0		# Set count to zero
rstrlend:
        lw $t0, 0($sp)		# Reload $t0
	lw $ra, 20($sp)		# Reload our return address
	addu $sp, $sp, 24	# Destroy our stack frame, and
	jr $ra			# return back to our caller


###################################################################
#			     strcmp()				  #
###################################################################

# Compare two strings and return 0 if they are same, -1 if the first
# comes first, 1 if the second one comes first. Error check the
# inputs and return -2 on error.
# Arguments: $a0, $a1 are pointers to the beginning of the strings
# Results: $v0 is the result: -1, 0, +1, or -2 on error
#
strcmp: subu $sp, $sp, 24       # Build the stack frame
        sw $ra, 20($sp)
        sw $t0, 0($sp)		# Save $t0 as we use it
        sw $t1, 4($sp)		# Save $t1 as we use it

	beqz $a0, cmperr	# Return -2 if either arg is NULL
	beqz $a1, cmperr	# Return -2 if either arg is NULL

cmploop:
	lb $t0, ($a0)		# Get char from string1
	lb $t1, ($a1)		# Get char from string2

	bge $t0, $t1, notlt	# $t0 isn't < $t1
	li $v0, -1		# $t0 was < $t1, so return -1
	b cmpend

notlt:	bge $t1, $t0, notlt2    # $t1 isn't < $t2
	li $v0, 1		# $t1 was < $t0, so return 1
	b cmpend

# At this point, we can deduce that the two characters
# are equal, so we should bump up the pointers and loop
# back. However, they could both be NUL, so we need to
# stop and return 0 if either of them are NULL

notlt2:	bnez $t0, notnull
	li $v0, 0		# Both strings ended, so must be equal
	b cmpend

notnull: addi $a0, $a0, 1	# Move pointers up to next char
	addi $a1, $a1, 1
	b cmploop		# Branch back if both chars were not NUL

cmperr:
	li $v0, -2		# Err, set result= -2
cmpend:
        lw $t0, 0($sp)		# Reload $t0
        lw $t1, 4($sp)		# Reload $t1
	lw $ra, 20($sp)		# Reload our return address
	addu $sp, $sp, 24	# Destroy our stack frame, and
	jr $ra			# return back to our caller


###################################################################
#			     memcmp()				  #
###################################################################
#
# This initial version isn't optimised as it does byte copying.
# However, it does try to solve the overlap problem.
# If the destination address > src address, then it's possible
# that the src end overlaps the dest beginning, so we copy from
# src end backwards down to src start. Alternatively,
# if the destination address < src address, then it's possible
# that the src beginning overlaps the dest end, so we copy from
# src beginning upwards to src end.
#
# Arguments: $a0, $a1 are pointers to the destination and src buffers,
# 	     $a2 is the length to be copied
# Results: $v0 is a copy of the destination pointer
#
memcpy: subu $sp, $sp, 24       # Build the stack frame
        sw $ra, 20($sp)
        sw $a0, 0($sp)		# Save the destination ptr for later
        sw $a1, 4($sp)		# Save the src ptr for later
        sw $a2, 8($sp)		# Save the count for later
        sw $t1, 12($sp)		# Save $t1 as we use it

	blez $a2, memend	# Error if count is <= 0
	beqz $a0, memend	# Error if dest pointer is NULL
	beqz $a1, memend	# Error if src pointer is NULL

	bgt $a0, $a1, backcopy	# Do the backwards copy if dest > src
	
fwdloop: 
	lb $t1, ($a1)		# Load byte from $a1
	sb $t1, ($a0)		# Save into destination
	add $a0, $a0, 1		# Move src and dst pointer up
	add $a1, $a1, 1
	sub $a2, $a2, 1		# And count down
	bnez $a2, fwdloop	# Loop back if more to copy

	b memend		# Exit the function
	
backcopy:
	# Before we start, add $a2 to both src and dest pointers.
	# But first decrement 1, then add it back on!
	sub $a2, $a2, 1
	add $a0, $a0, $a2
	add $a1, $a1, $a2
	add $a2, $a2, 1
	
backloop:
	lb $t1, ($a1)		# Load byte from $a1
	sb $t1, ($a0)		# Save into destination
	sub $a0, $a0, 1		# Move src and dst pointer down
	sub $a1, $a1, 1
	sub $a2, $a2, 1		# And count down
	bnez $a2, backloop	# Loop back if more to copy

memend: 
        lw $a0, 0($sp)		# Reload the destination register
        lw $a1, 4($sp)		# Reload the src register
        lw $a2, 8($sp)		# Reload the count register
        lw $t1, 12($sp)		# Reload $t1
	move $v0, $a0		# Set the result value
	lw $ra, 20($sp)		# Reload our return address
	addu $sp, $sp, 24	# Destroy our stack frame, and
	jr $ra			# return back to our caller

###################################################################
#			     strdup()				  #
###################################################################
#
# Given $a0 as a pointer to a string, duplicate the string and
# return a pointer to the duplicate. We need to malloc a new
# buffer which is the size of the existing string (including the
# NUL character), then memcpy() the old string into the new string.
# If we are given a NULL pointer, we simply return NULL. Similarly,
# if something goes wrong with malloc() or strlen(), we return NULL.
#
strdup:	subu $sp, $sp, 24       # Build the stack frame
        sw $ra, 20($sp)
        sw $a0, 0($sp)		# Save the string pointer for later

	beqz $a0, duperr	# Error if string pointer is NULL

	# Get the length of the string into $v0
	# Increment it to account for the NUL character
	jal strlen
	add $v0, $v0, 1

	# Malloc that much memory
	move $a0, $v0
	jal malloc
	beqz $v0, duperr	# Error if malloc returns NULL

	# At this point, $a0 is the length, $v0 is the dest pointer
	# and the stack has the src pointer. Move them around
	# so that we can do the memcpy
	move $a2, $a0		# Set the length
	move $a0, $v0		# Set the dest pointer
        lw $a1, 0($sp)		# Reload the src pointer
	jal memcpy		# Do the copy
	b dupend		# and return with the $v0 result

duperr:	la $v0, 0		# Return NULL if error
dupend:
	lw $ra, 20($sp)		# Reload our return address
	addu $sp, $sp, 24	# Destroy our stack frame, and
	jr $ra			# return back to our caller
